<?php

namespace Drupal\opigno_learning_path;

use Drupal\Core\Entity\EntityFormBuilderInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\group\Entity\GroupContent;

/**
 * Class JoinService.
 */
class JoinService {

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The entity form builder.
   *
   * @var \Drupal\Core\Entity\EntityFormBuilderInterface
   */
  protected $entityFormBuilder;

  /**
   * Constructs a new JoinService object.
   */
  public function __construct(AccountInterface $current_user, EntityFormBuilderInterface $entity_form_builder) {
    $this->currentUser = $current_user;
    $this->entityFormBuilder = $entity_form_builder;
  }

  /**
   * {@inheritdoc}
   */
  public function getForm($group) {
    // Show a join form, or a request membership form depending on group
    // validation settings.
    $learning_path_id = $group->id();
    $requires_validation = $group->requiresValidation();
    if ($requires_validation) {
      $plugin = $group->getGroupType()->getContentPlugin('group_membership_request');
      $form_name = 'group-request-membership';
    }
    else {

      $classes = $group->getClasses();
      if (\count($classes) === 0) {
        return [];
      }
      $group = \reset($classes);
      $plugin = $group->getGroupType()->getContentPlugin('group_membership');
      $form_name = 'group-join';
    }

    // Pre-populate group content with the current user.
    $group_content = GroupContent::create([
      'type' => $plugin->getContentTypeConfigId(),
      'gid' => $group->id(),
      'entity_id' => $this->currentUser->id(),
    ]);
    if ($requires_validation) {
      $group_content->set('grequest_status', 'pending');
    }
    $group_content->learning_path_id = $learning_path_id;

    $form = $this->entityFormBuilder->getForm($group_content, $form_name, ['learning_path_id' => $learning_path_id]);
    return $form;
  }

}
